package ru.dbotthepony.mc.otm.block.entity.storage

import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.StringTag
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.network.chat.Component
import net.minecraft.network.chat.TranslatableComponent
import net.minecraft.server.level.ServerLevel
import net.minecraft.server.level.ServerPlayer
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.state.BlockState
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.util.INBTSerializable
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.network.NetworkEvent
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.block.entity.MatteryPoweredBlockEntity
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.WorkerEnergyStorage
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.graph.storage.BasicStorageGraphNode
import ru.dbotthepony.mc.otm.graph.storage.StorageNetworkGraph
import ru.dbotthepony.mc.otm.ifHas
import ru.dbotthepony.mc.otm.menu.ItemMonitorMenu
import ru.dbotthepony.mc.otm.registry.MBlockEntities
import ru.dbotthepony.mc.otm.set
import java.util.UUID
import java.util.function.Supplier

class ItemMonitorPlayerSettings : INBTSerializable<CompoundTag> {
	enum class MatchMode {
		FUZZY,
		EXACT
	}

	enum class Target {
		SYSTEM,
		INVENTORY
	}

	enum class Amount {
		ONE,
		STACK,
		FULL
	}

	var ingredientMatchingMode: MatchMode = MatchMode.FUZZY
	var ingredientPriority = Target.SYSTEM
	var resultTarget = Target.INVENTORY
	var craftingAmount = Amount.STACK

	override fun serializeNBT(): CompoundTag {
		return CompoundTag().also {
			it["ingredientMatchingMode"] = ingredientMatchingMode.name
			it["ingredientPriority"] = ingredientPriority.name
			it["resultTarget"] = resultTarget.name
			it["craftingAmount"] = craftingAmount.name
		}
	}

	override fun deserializeNBT(nbt: CompoundTag) {
		nbt.ifHas("ingredientMatchingMode", StringTag::class.java) { ingredientMatchingMode = MatchMode.valueOf(it.asString) }
		nbt.ifHas("ingredientPriority", StringTag::class.java) { ingredientPriority = Target.valueOf(it.asString) }
		nbt.ifHas("resultTarget", StringTag::class.java) { resultTarget = Target.valueOf(it.asString) }
		nbt.ifHas("craftingAmount", StringTag::class.java) { craftingAmount = Amount.valueOf(it.asString) }
	}

	fun read(buff: FriendlyByteBuf) {
		ingredientMatchingMode = buff.readEnum(MatchMode::class.java)
		ingredientPriority = buff.readEnum(Target::class.java)
		resultTarget = buff.readEnum(Target::class.java)
		craftingAmount = buff.readEnum(Amount::class.java)
	}

	fun read(other: ItemMonitorPlayerSettings) {
		ingredientMatchingMode = other.ingredientMatchingMode
		ingredientPriority = other.ingredientPriority
		resultTarget = other.resultTarget
		craftingAmount = other.craftingAmount
	}

	fun write(buff: FriendlyByteBuf) {
		buff.writeEnum(ingredientMatchingMode)
		buff.writeEnum(ingredientPriority)
		buff.writeEnum(resultTarget)
		buff.writeEnum(craftingAmount)
	}

	private fun playClient() {
		val ply = minecraft.player ?: throw IllegalStateException("Player is missing")
		val container = ply.containerMenu as? ItemMonitorMenu ?: return LOGGER.error("Received ItemMonitorPlayerSettings but container is missing or not of right type ({})!", ply.containerMenu)

		container.settings.read(this)
	}

	private fun playServer(ply: ServerPlayer) {
		val container = ply.containerMenu as? ItemMonitorMenu ?: return LOGGER.error("Received ItemMonitorPlayerSettings from {} but container is missing or not of right type ({})!", ply, ply.containerMenu)
		container.settings.read(this)
		(container.tile as ItemMonitorBlockEntity).setChangedLight()
	}

	fun play(context: Supplier<NetworkEvent.Context>) {
		context.get().packetHandled = true
		val ply = context.get().sender

		if (ply != null) {
			playServer(ply)
		} else {
			playClient()
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf): ItemMonitorPlayerSettings {
			return ItemMonitorPlayerSettings().also { it.read(buff) }
		}

		private val LOGGER = LogManager.getLogger()
	}
}

class ItemMonitorBlockEntity(p_155229_: BlockPos, p_155230_: BlockState) :
	MatteryPoweredBlockEntity(MBlockEntities.ITEM_MONITOR, p_155229_, p_155230_) {

	override val energy = WorkerEnergyStorage(this)
	val cell = BasicStorageGraphNode(energy)

	val settings = HashMap<UUID, ItemMonitorPlayerSettings>()

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)

		nbt.put("player_settings", CompoundTag().also {
			for ((key, value) in this.settings) {
				it[key.toString()] = value.serializeNBT()
			}
		})
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)
		this.settings.clear()

		val settings = nbt.getCompound("player_settings")

		for (key in settings.allKeys) {
			val uuid = UUID.fromString(key)
			val deserialized = ItemMonitorPlayerSettings()
			deserialized.deserializeNBT(settings.getCompound(key))

			check(this.settings.put(uuid, deserialized) == null) { "Duplicate UUID??? $uuid" }
		}
	}

	fun getSettings(ply: ServerPlayer): ItemMonitorPlayerSettings {
		return settings.computeIfAbsent(ply.uuid) { ItemMonitorPlayerSettings() }
	}

	fun tick() {
		batteryChargeLoop()
		cell.tickEnergyDemanding()
	}

	override val defaultDisplayName: Component
		get() = MACHINE_NAME

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return ItemMonitorMenu(containerID, inventory, this)
	}

	override fun setLevel(p_155231_: Level) {
		super.setLevel(p_155231_)

		if (p_155231_ is ServerLevel)
			StorageNetworkGraph.discoverFull(this, cell.storageNode)
	}

	override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		return if (cap === MatteryCapability.STORAGE_NODE) {
			cell.get().cast()
		} else super.getCapability(cap, side)
	}

	override fun invalidateCaps() {
		super.invalidateCaps()
		cell.invalidate()
	}

	override fun reviveCaps() {
		super.reviveCaps()
		cell.revive()
	}

	override fun setRemoved() {
		super.setRemoved()
		cell.destroy(level)
	}

	companion object {
		private val MACHINE_NAME = TranslatableComponent("block.overdrive_that_matters.item_monitor")
	}
}